/* This file is part of the wvWare 2 project
   Copyright (C) 2001 Werner Trobin <trobin@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "olestream.h"
#include <stdio.h>
#include <iostream>

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

using namespace wvWare;

OLEStream::OLEStream( MsOleStream* stream, OLEStorage* storage ) :
    m_stream( stream ), m_storage( storage )
{
}

OLEStream::~OLEStream()
{
    if ( !m_stream ) {
        std::cerr << "OLEStream::~OLEStream(): Double deletion detected" << std::endl;
        return;
    }
    if ( m_storage )
        m_storage->streamDestroyed( this );
    ms_ole_stream_close( &m_stream );
    m_stream = 0;
    m_storage = 0;
}

int OLEStream::seek( int offset, int whence )
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return -1;
#endif
    return m_stream->lseek( m_stream, offset, static_cast<MsOleSeek>( whence ) );
}

int OLEStream::tell() const
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return -1;
#endif
    return m_stream->tell( m_stream );
}

size_t OLEStream::size() const
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return 0;
#endif
    return m_stream->size;
}

void OLEStream::push()
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return;
#endif
    m_positions.push( tell() );
}

bool OLEStream::pop()
{
    if ( !m_stream || m_positions.empty() )
        return false;
    seek( m_positions.top(), SEEK_SET );
    m_positions.pop();
    return true;
}


OLEStreamReader::OLEStreamReader( MsOleStream* stream, OLEStorage* storage ) :
    OLEStream( stream, storage )
{
}

OLEStreamReader::~OLEStreamReader()
{
}

U8 OLEStreamReader::readU8()
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return 0;
#endif

    U8 ret;
    m_stream->read_copy( m_stream, static_cast<guint8*>( &ret ), 1 );
    return ret;
}

S8 OLEStreamReader::readS8()
{
    return static_cast<S8>( readU8() );
}

U16 OLEStreamReader::readU16()
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return 0;
#endif

#if defined(WORDS_BIGENDIAN)
    // Please take care when "optimizing" that and read
    // http://gcc.gnu.org/ml/gcc-bugs/2000-12/msg00429.html
    // and http://www.eskimo.com/~scs/C-faq/q3.8.html
    U16 tmp1 = readU8();
    U16 tmp2 = readU8();
    return ( tmp2 << 8 ) | tmp1;
#else
    U16 ret;
    m_stream->read_copy( m_stream, reinterpret_cast<guint8*>( &ret ), 2 );
    return ret;
#endif
}

S16 OLEStreamReader::readS16()
{
    return static_cast<S16>( readU16() );
}

U32 OLEStreamReader::readU32()
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return 0;
#endif

#if defined(WORDS_BIGENDIAN)
    // Please take care when "optimizing" that and read
    // http://gcc.gnu.org/ml/gcc-bugs/2000-12/msg00429.html
    // and http://www.eskimo.com/~scs/C-faq/q3.8.html
    U32 tmp1 = readU16();
    U32 tmp2 = readU16();
    return ( tmp2 << 16 ) | tmp1;
#else
    U32 ret;
    m_stream->read_copy(m_stream, reinterpret_cast<guint8*>( &ret ), 4 );
    return ret;
#endif
}

S32 OLEStreamReader::readS32()
{
    return static_cast<S32>( readU32() );
}

bool OLEStreamReader::read( U8 *buffer, size_t lenght )
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return false;
#endif
    return m_stream->read_copy( m_stream, buffer, lenght ) != 0;
}

void OLEStreamReader::dumpStream( const std::string& fileName )
{
    push();
    seek( 0, SEEK_SET );

    FILE* myFile = fopen( fileName.c_str(), "w" );
    if ( !myFile ) {
        pop();
        return;
    }

    const size_t buflen = 1024;
    char buffer[ buflen ];
    size_t remaining = size();
    size_t length;

    while ( remaining ) {
        length = remaining > buflen ? buflen : remaining;
        remaining -= length;
        m_stream->read_copy( m_stream, reinterpret_cast<guint8*>( buffer ), length );
        fwrite( buffer, 1, length, myFile );
    }

    fclose( myFile );
    pop();
}


OLEStreamWriter::OLEStreamWriter( MsOleStream* stream, OLEStorage* storage ) :
    OLEStream( stream, storage )
{
}

OLEStreamWriter::~OLEStreamWriter()
{
}

void OLEStreamWriter::write( U8 data )
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return;
#endif
    m_stream->write( m_stream, &data, 1 );
}

void OLEStreamWriter::write( S8 data )
{
    write( static_cast<U8>( data ) );
}

void OLEStreamWriter::write( U16 data )
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return;
#endif
    U16 copy = toLittleEndian( data );
    m_stream->write( m_stream, reinterpret_cast<guint8 *>( &copy ), 2 );
}

void OLEStreamWriter::write( S16 data )
{
    write( static_cast<U16>( data ) );
}

void OLEStreamWriter::write( U32 data )
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return;
#endif
    U32 copy = toLittleEndian( data );
    m_stream->write( m_stream, reinterpret_cast<guint8 *>( &copy ), 4 );
}

void OLEStreamWriter::write( S32 data )
{
    write( static_cast<U32>( data ) );
}

void OLEStreamWriter::write( U8* data, size_t length )
{
#ifdef WV2_CHECKING
    if ( !m_stream )
        return;
#endif
    m_stream->write( m_stream, data, length );
}
