/* This file is part of the wvWare 2 project
   Copyright (C) 2001 Werner Trobin <trobin@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#ifndef PARSER97_H
#define PARSER97_H

#include "parser.h"
#include "parser9x.h"
#include "word97_helper.h"

#include <list>
#include <stack>

namespace wvWare
{

class StyleSheet;
class Properties97;
class FontCollection;
class ListInfoProvider;
class TextConverter;
class Headers97;
class Footnotes97;
class Fields;

// Helper structures for the Functor-based approach
class HeaderData;
class FootnoteData;
class TableRowData;

/**
 * This class is the main parser class for Word97 documents.
 */
class Parser97 : public Parser, private Parser9x
{
public:
    Parser97( OLEStorage* storage, OLEStreamReader* wordDocument );
    virtual ~Parser97();

    /**
     * The main parsing method
     */
    virtual bool parse();

    virtual const Word97::FIB& fib() const;
    virtual const Word97::DOP& dop() const;

    /**
     * Get the font family name structure for a given ftc.
     */
    virtual const Word97::FFN& font( S16 ftc ) const;

    /**
     * Get the associated strings (author, title,...).
     * Not cached.
     */
    virtual AssociatedStrings associatedStrings();

    virtual const StyleSheet& styleSheet() const;


    // This part of the public API is only visible to the Functor classes,
    // as the "outside world" only sees the public API of Parser. The Functors
    // allow to delay the parsing of certain text inside the file (e.g. headers)
    // and trigger parsing at any point (as long as the parser exists).
    //
    // In case you want to add a new method here, please obey the following guidelines:
    //    - Executing the method mustn't change the state of the parser
    //    - Be very careful, these calls can possibly be triggered at any time
    void parseHeaders( const HeaderData& data );
    void parseFootnote( const FootnoteData& data );
    void parseTableRow( const TableRowData& data );

private:
    // Uniquely represents a position inside a complex file. Used to map a CP to a Position
    struct Position
    {
        // Start position
        Position( U32 p, U32 o ) : piece( p ), offset( o ) {}
         // Constructs a Position from a CP
        Position( U32 cp, const Word97::PLCF<Word97::PCD>* plcfpcd );

        U32 piece;    // The piece number (0-based index)
        U32 offset;   // The CP offset within the piece
    };

    // Represents a chunk of text. This is a part of a (or a whole) paragraph
    // contained in one text piece. A paragraph consists of at least one Chunk.
    // We don't store the paragraph/section mark, and in case only the paragraph
    // mark sits in a different piece than the rest of the paragraph we just store
    // an empty string for this chunk.
    struct Chunk
    {
        Chunk( const UString& text, const Position& position, U32 startFC, bool isUnicode ) :
            m_text( text ), m_position( position ), m_startFC( startFC ), m_isUnicode( isUnicode ) {}

        UString m_text;
        Position m_position;
        U32 m_startFC;
        bool m_isUnicode;
    };
    // Represents a paragraph consisting of at least one Chunk. Right now it's only
    // a typedef, maybe we need more than that later on
    typedef std::list<Chunk> Paragraph;

    // We have to keep track of the current parsing mode (e.g. are we skimming tables
    // or are we parsing them?)
    enum ParsingMode { Default, Table };

    // don't copy or assign us
    Parser97( const Parser97& rhs );
    Parser97& operator=( const Parser97& rhs );

    void init();
    bool readPieceTable();

    bool parseBody();

    // Expects m_remainingChars to be set correctly, changes the state of m_wordDocument,...
    void parseHelper( Position startPos );
    template<typename String> void processPiece( String* string, U32 fc, U32 limit, const Position& position );
    // These helper methods are a cheap trick to "configure" parts of the template code by
    // plain old overloading. It's just a matter of compressed vs. real unicode (1 vs. 2 bytes)
    UString processPieceStringHelper( XCHAR* string, unsigned int start, unsigned int index ) const;
    UString processPieceStringHelper( U8* string, unsigned int start, unsigned int index ) const;
    // Processes the current contents of the Paragraph structure and clears it when it's done
    void processParagraph( U32 fc );
    void processChunk( const Chunk& chunk, SharedPtr<const Word97::CHP> chp,
                       U32 length, U32 index, U32 currentStart );
    void processRun( const Chunk& chunk, SharedPtr<const Word97::CHP> chp,
                     U32 length, U32 index, U32 currentStart );

    void processSpecialCharacter( UChar character, U32 globalCP, SharedPtr<const Word97::CHP> chp );
    void processFootnote( UChar character, U32 globalCP, SharedPtr<const Word97::CHP> chp );

    // Helper methods to gather and emit the information needed for the functors
    void emitHeaderData( SharedPtr<const Word97::SEP> sep );

    void parseHeader( const HeaderData& data, unsigned char mask );

    void saveState( U32 newRemainingChars, SubDocument newSubDocument, ParsingMode newParsingMode = Default );
    void restoreState();

    // Maps the global CP (as found in the piece table) to the local CP
    // coordinate space of the corresponding sub document
    U32 toLocalCP( U32 globalCP ) const;
    // Calculates the real FC and tells us whether it was unicode or not
    inline void realFC( U32& fc, bool& unicode ) const;
    // Helper method to use std::accumulate in the table handling code
    static int accumulativeLength( int len, const Chunk& chunk );

    // First all variables which don't change their state during
    // the parsing process. We don't have to save and restore those.
    OLEStreamReader* m_table;        // table stream ('[01]Table')
    OLEStreamReader* m_data;         // data stream ('Data'), might be 0

    Properties97* m_properties;
    FontCollection* m_fonts;
    ListInfoProvider* m_lists;
    Footnotes97* m_footnotes;
    Headers97* m_headers;
    Fields* m_fields;

    const Word97::FIB m_fib;
    Word97::PLCF<Word97::PCD>* m_plcfpcd;     // piece table

    TextConverter* m_textconverter;

    // From here on we have all variables which change their state depending
    // on the parsed content. These variables have to be saved and restored
    // to make the parsing code reentrant.
    Position* m_tableRowStart;      // If != 0 this represents the start of a table row
    U32 m_tableRowLength;           // Lenght of the table row (in characters). Only valid
    bool m_cellMarkFound;           // if m_tableRowStart != 0
    int m_remainingCells;           // The number of remaining cells for the processed row

    Paragraph* m_currentParagraph;

    U32 m_remainingChars;
    U32 m_sectionNumber;

    // Keeps track of the current sub document
    SubDocument m_subDocument;

    // We have to behave differently, depending whether we are parsing
    // a table or the "main" text, as we skim the table first
    ParsingMode m_parsingMode;

    // Needed to have reentrant parsing methods (to make the functor approach work)
    struct ParsingState
    {
        ParsingState( Position* tableRowS, U32 tableRowL, int remCells, Paragraph* parag,
                      U32 remChars, U32 sectionNum, SubDocument subD, ParsingMode mode ) :
            tableRowStart( tableRowS ), tableRowLength( tableRowL ), remainingCells( remCells ),
            paragraph( parag ), remainingChars( remChars ), sectionNumber( sectionNum ),
            subDocument( subD ), parsingMode( mode ) {}

        Position* tableRowStart;
        U32 tableRowLength;
        int remainingCells;
        Paragraph* paragraph;
        U32 remainingChars;
        U32 sectionNumber;   // not strictly necessary, but doesn't hurt
        SubDocument subDocument;
        ParsingMode parsingMode;
    };

    std::stack<ParsingState> oldParsingStates;
};

inline void Parser97::realFC( U32& fc, bool& unicode ) const
{
    if ( fc & 0x40000000 ) {
        fc = ( fc & 0xbfffffff ) >> 1;
        unicode = false;
    }
    else
        unicode = true;
}

} // namespace wvWare

#endif // PARSER97_H
