/* This file is part of the wvWare 2 project
   Copyright (C) 2002 Werner Trobin <trobin@kde.org>

   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Library General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Library General Public License for more details.

   You should have received a copy of the GNU Library General Public License
   along with this library; see the file COPYING.LIB.  If not, write to
   the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
   Boston, MA 02111-1307, USA.
*/

#include "handlertest.h"
#include <stdexcept>

#include <test.h>
#include <parser.h>
#include <parserfactory.h>
#include <word97_generated.h>
#include <paragraphproperties.h>

using namespace wvWare;

// This file contains a small test harness for functional tests of the
// wv2 API. It has two modes of operation: recording mode and testing mode.
// When using the recording mode you can "learn" the correct order of
// callbacks in a file, and testing mode uses the previously recorded file
// to detect differences after changes in the codebase.

IOManager::IOManager( const std::string& vignette, Mode mode ) :
    m_mode( mode ), m_vignette( 0 ), m_buffer( 0 ),
    m_bufindex( 0 ), m_buflen( 0 ), m_isOk( true )
{
    m_vignette = fopen( vignette.c_str(), mode == Record ? "w" : "r" );
    if ( !m_vignette )
        m_isOk = false;
    else {
        if ( mode == Verify )
            m_buffer = new char[ BUFLEN ];
    }
}

IOManager::~IOManager()
{
    delete [] m_buffer;
    fclose( m_vignette );
}

bool IOManager::atEOF() const
{
    if ( !m_vignette )
        return false;
    return feof( m_vignette );
}

void IOManager::process( const char* string )
{
    if ( m_mode == Record ) {
        fwrite( string, sizeof( char ), strlen( string ), m_vignette );
        fwrite( "\n", sizeof( char ), 1, m_vignette );
    }
    else {
        std::string line( getNextLine() );
        if ( line != string )
            std::cerr << "Error: Got '" << string << "', expected '" << line << "'" << std::endl;
    }
}

std::string IOManager::getNextLine()
{
    std::string s;

    do {
        readFromBuffer( s );
    } while ( m_bufindex >= m_buflen && fillBuffer() );

    ++m_bufindex;
    return s;
}

bool IOManager::fillBuffer()
{
    m_buflen = fread( m_buffer, 1, BUFLEN, m_vignette );
    m_bufindex = 0;
    return m_buflen != 0 || feof( m_vignette ) == 0;
}

void IOManager::readFromBuffer( std::string& s )
{
    for ( ; m_bufindex < m_buflen && m_buffer[ m_bufindex ] != '\n'; ++m_bufindex )
        s += m_buffer[ m_bufindex ];
}


InlineReplacementTest::InlineReplacementTest( IOManager* iomanager ) :
    m_iomanager( iomanager )
{
}

U8 InlineReplacementTest::tab()
{
    m_iomanager->process( "INLINE: tab" );
    return InlineReplacementHandler::tab();
}

U8 InlineReplacementTest::hardLineBreak()
{
    m_iomanager->process( "INLINE: hard line break" );
    return InlineReplacementHandler::hardLineBreak();
}

U8 InlineReplacementTest::nonBreakingHyphen()
{
    m_iomanager->process( "INLINE: non breaking hyphen" );
    return InlineReplacementHandler::nonBreakingHyphen();
}

U8 InlineReplacementTest::nonRequiredHyphen()
{
    m_iomanager->process( "INLINE: non required hyphen" );
    return InlineReplacementHandler::nonRequiredHyphen();
}

U8 InlineReplacementTest::nonBreakingSpace()
{
    m_iomanager->process( "INLINE: non breaking space" );
    return InlineReplacementHandler::nonBreakingSpace();
}


TextTest::TextTest( IOManager* iomanager ) :
    m_iomanager( iomanager )
{
}

void TextTest::sectionStart( SharedPtr<const Word97::SEP> sep )
{
    m_iomanager->process( "TEXT: section start" );
    TextHandler::sectionStart( sep );
}

void TextTest::sectionEnd()
{
    m_iomanager->process( "TEXT: section end" );
    TextHandler::sectionEnd();
}

void TextTest::pageBreak()
{
    m_iomanager->process( "TEXT: page break" );
    TextHandler::pageBreak();
}

void TextTest::headersFound( const HeaderFunctor& parseHeaders )
{
    m_iomanager->process( "TEXT: headers found" );
    TextHandler::headersFound( parseHeaders );
}

void TextTest::paragraphStart( SharedPtr<const ParagraphProperties> paragraphProperties )
{
    m_iomanager->process( "TEXT: paragraph start" );
    TextHandler::paragraphStart( paragraphProperties );
}

void TextTest::paragraphEnd()
{
    m_iomanager->process( "TEXT: paragraph end" );
    TextHandler::paragraphEnd();
}

void TextTest::runOfText( const UString& text, SharedPtr<const Word97::CHP> chp )
{
    m_iomanager->process( "TEXT: run of text" );
    TextHandler::runOfText( text, chp );
}

void TextTest::specialCharacter( SpecialCharacter character, SharedPtr<const Word97::CHP> chp )
{
    m_iomanager->process( "TEXT: special character" );
    TextHandler::specialCharacter( character, chp );
}

void TextTest::footnoteFound( FootnoteData::Type type, UChar character,
                              SharedPtr<const Word97::CHP> chp, const FootnoteFunctor& parseFootnote )
{
    m_iomanager->process( "TEXT: footnote found" );
    TextHandler::footnoteFound( type, character, chp, parseFootnote );
}

void TextTest::footnoteAutoNumber( SharedPtr<const Word97::CHP> chp )
{
    m_iomanager->process( "TEXT: footnote auto-number" );
    TextHandler::footnoteAutoNumber( chp );
}

void TextTest::fieldStart( const FLD* fld, SharedPtr<const Word97::CHP> chp )
{
    m_iomanager->process( "TEXT: field start" );
    TextHandler::fieldStart( fld, chp );
}

void TextTest::fieldSeparator( const FLD* fld, SharedPtr<const Word97::CHP> chp )
{
    m_iomanager->process( "TEXT: field separator" );
    TextHandler::fieldSeparator( fld, chp );
}

void TextTest::fieldEnd( const FLD* fld, SharedPtr<const Word97::CHP> chp )
{
    m_iomanager->process( "TEXT: field end" );
    TextHandler::fieldEnd( fld, chp );
}


int main( int argc, char** argv )
{
    if ( argc != 4 ) {
        std::cerr << "Usage: handlertest foo.doc -r foo.rec for recording\n"
                  << "       handlertest foo.doc -v foo.rec to verify" << std::endl;
        ::exit( 1 );
    }

    std::string document( argv[ 1 ] );
    IOManager::Mode mode = IOManager::Verify;
    if ( strcmp( argv[ 2 ], "-r" ) == 0 )
        mode = IOManager::Record;

    SharedPtr<Parser> parser( ParserFactory::createParser( document ) );
    if ( !parser ) {
        std::cerr << "Error: Couldn't create a parser for this document" << std::endl;
        ::exit( 2 );
    }

    IOManager* iomanager = new IOManager( argv[ 3 ], mode );
    if ( !iomanager->isOk() ) {
        std::cerr << "Error: Couldn't open the recorded file" << std::endl;
        delete iomanager;
        ::exit( 3 );
    }

    TextTest* textTest = new TextTest( iomanager );
    parser->setTextHandler( textTest );

    InlineReplacementTest* replacementTest = new InlineReplacementTest( iomanager );
    parser->setInlineReplacementHandler( replacementTest );

    parser->parse();

    if ( mode == IOManager::Verify && !iomanager->atEOF() ) {
        std::cerr << "Error: The file is larger than I expected" << std::endl;
        delete replacementTest;
        delete textTest;
        delete iomanager;
        ::exit( 4 );
    }
    std::cout << "Done." << std::endl;

    delete replacementTest;
    delete textTest;
    delete iomanager;

    return 0;
}
